// filter.cpp
// (c) 2008 Pedro Fortuny Ayuso
// See http://pfortuny.net/filter.html for the details
// This software is free, it is probably useless to you, but you
// may even learn something. In any case, it is distributed with
// NO GUARANTEE whatsoever concerning any intended or unintended
// result. You are free to distribute, modify, copy, sell and
// buy it, but you are not allowed to stand as the original author.
#include <algorithm>
#include <fstream>
#include <iomanip>
#include <iostream>
#include <iterator>
#include <list>
#include <map>
#include <set>
#include <sstream>
#include <string>
#include <queue>
#include <vector>

std::string accents   = "";
std::string unaccents = "AEIOUAEIOU";
std::string small_caps = "0123456789abcdefghijklmnopqrstuvwxyz";
std::string alnum      = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZAEIOUAEIOU";
std::string accented   = "";
std::string all_valid  = small_caps + alnum + accented;
std::map<char, char> upper;  // defined in main (maps char to CHAR)


void to_up(std::string &word){

        for(size_t i = 0;
            i < word.size();
            i++){
                word[i] = upper[word[i]];
        }

        return;
}


void add_words(int counter, 
               std::string line, 
               std::map<std::string, std::vector<int> > &Bwords){


        std::vector<std::string> words;
        size_t start = 0;
        size_t end   = 0;
        // split line into words
        while((start = line.find_first_of(all_valid, start)) != line.npos){
                end = line.find_first_not_of(all_valid, start);
                std::string word = line.substr(start, end - start);
                to_up(word);
                words.push_back(word);
                start = end;
                Bwords[word].push_back(counter);
        }

        return;
}



int main(int argc, char * argv[]){

        std::map<std::string, std::vector<int> > Bwords;
        
        std::ifstream list("sentences.txt");
        std::string current_line;

        for(size_t i = 0;
            i < small_caps.size();
            i++){
                upper[small_caps[i]] = alnum[i];
                // DON'T FORGET THESE!!! (we are NOT searching in to_up),
                // just replacing
                upper[alnum[i]]      = alnum[i];
        }

        // read first spurious line
        getline(list, current_line, '\n');

        // run through the first list really fast
        do {
                getline(list, current_line, '\n');	
        } while(current_line.compare("SECOND LIST") != 0);

        // parse the second list; why are we doing this?
        // because FOR THIS SPECIFIC PROBLEM, it is simpler having
        // a hash 
        // word -> lines
        // for the 'words' in the second list and then matching
        // looking for the specific matches, line by line in the
        // first list.
        // This is probably a matter of opinion, but I gues I'm right...

        int B_counter = 1;
        // parse the second list
        do {
                getline(list, current_line, '\n');
                add_words(B_counter, current_line, Bwords);
                B_counter++ ;

        } while(list.eof() != true);

        // list.seekg(0,std::ios_base::beg) does not work for some reason...
        // nor does close() and open(), one needs a new ifstream. STRANGE
        // (this happens on a NetBSD Alpha-64 with gcc-3)
        list.close();

        std::ifstream list2("sentences.txt");
        // omit first spurious line ('FIRST LIST')
        getline(list2, current_line, '\n');        
        getline(list2, current_line, '\n');
        int A_counter = 1;

        // parse line by line only the first list
        // and do the real work
        while(current_line.compare("SECOND LIST") != 0){
                std::vector<std::string> words_in_line;
                size_t start = 0;
                size_t end   = 0;
                // make a vector of the upcase words in line
                while(end != 
                      current_line.npos){
                        end = current_line.find_first_of(" ", start);
                        std::string word = current_line.substr(start, end - start);
                        to_up(word);
                        words_in_line.push_back(word);
                        start = end + 1;
                }

                // If line is w1 w2 ... wk (k words), 
                // compute the intersection of Bwords[w1], Bwords[w2], ..., Bwords[wk]
                // those are the lines in SECOND LIST containing ALL the words in the
                // present line of FIRST LIST
                std::vector<int> inter(Bwords[words_in_line[0]]);
                // tmp(0) ends up in a segfault on some architectures... funny, isn't it?
                std::vector<int> tmp(1);
                std::vector<int>::iterator zero = Bwords[words_in_line[0]].begin();
                std::vector<int>::iterator infinity = Bwords[words_in_line[0]].end();
                for(size_t i = 1;
                    i < words_in_line.size();
                    i++){
                        // Bwords[word] is already ordered (important for intersecting)
                        infinity = set_intersection(inter.begin(), 
                                                    inter.end(),
                                                    Bwords[words_in_line[i]].begin(), 
                                                    Bwords[words_in_line[i]].end(),
                                                    tmp.begin());
                        // Is this necessary? Might I use inter.begin() and inter.end()
                        // above instead of [tmp.begin(), inifity) Looks like I cannot.
                        // This shows the limitations inherent to the set_intersection
                        // algorithm (is not iterable, and it looks like none of the
                        // 'set' algorithms is).
                        inter = std::vector<int>(0);
                        for(std::vector<int>::iterator i = tmp.begin();
                            i < infinity;
                            i++) inter.push_back(*i);
                        tmp = std::vector<int>(0);
                }
                // print the result
                for(std::vector<int>::iterator i = inter.begin();
                    i < inter.end();
                    i++){
                        std::cout << "A" << A_counter << 
                                " B" << *i << std::endl;
                }
                ++A_counter;
                getline(list2, current_line, '\n');
        }       
        return(0);
}
